'MP3 Extractor v1.0
'Chapin Information Services, Inc.
'info-svc.com
'10/16/2006
'
'Description:  Utility quickly restores fragmented MP3 files.  This version looks for fixed-length fragments in MP3 files only.
'
'Concept:  Media encapsulation can be removed without knowing the exact format.
'
'Distribution:  Original files from .info-svc.com.  Private use only.  Free.  Not for resale.  Includes source code.
'
'Thanks:
'http://mpgedit.org/mpgedit/mpeg_format/mpeghdr.htm
'http://www.id3.org/id3v2.4.0-structure.txt
'http://msdn.microsoft.com/vstudio/express/vb/
'http://www.yellowpipe.com/yis/download/free-icons.php
'
'Known Bugs:
'  .NET Framework Initialization Error raised if not installed.
'  Security error appears if run from a network drive.
'
'History:
'  10/16/06 Build 4  Added VBR handling.  Fixed off-by-one bugs in array functions.
'  10/02/06 Build 3  Added more synchronization logic.
'  10/01/06 Build 2  VBR attempt in Build 1 caused problems.  Reverted some changes.
'  10/01/06 Build 1  Fixed memory issues.
'  10/01/06 Build 0

Option Explicit On

Module Extractor

    Public Sub ExtractV2(ByVal Path As String)

        On Error GoTo ErrorHandling

        Const InitialBufferSize As Long = 20000

        Dim ID3v2Begin() As Byte = {Asc("I"), Asc("D"), Asc("3")}
        Dim MP3FrameBeginCRC() As Byte = {&HFF, &HFA}
        Dim MP3FrameBeginNoCRC() As Byte = {&HFF, &HFB}

        Dim MP3FrameBegin(1) As Byte  'synch first two bytes
        Dim MP3FrameBegin2 As Byte    'synch fourth byte

        Dim FileInStream As IO.FileStream
        Dim FileOutStream As IO.FileStream

        Dim FileBuffer(InitialBufferSize - 1) As Byte

        Dim samplerate As Long = 0

        Dim CurrPos As Long
        Dim TempPos As Long
        Dim BeginPos As Long
        Dim FileSize As Long
        Dim FrameSize As Long
        Dim FirstDWord(3) As Byte
        Dim NewPath As String
        Dim ExtraDataFound As Boolean = False
        Dim ExtraInfoSize As Long
        Dim ExtraInfoInterval As Long
        Dim EndOfFile1 As Long
        Dim EndOfFile2 As Long

        On Error Resume Next
        FileSize = FileLen(Path)
        If Err.Number <> 0 Then
            Err.Clear()
            MsgBox("Source File Could Not Be Found", MsgBoxStyle.Information, "No Music For You")
            Exit Sub
        End If
        NewPath = Path & ".mp3"
        FileSize = FileLen(NewPath)
        If Err.Number = 0 Then
            MsgBox("Destination File Already Exists, Aborting!", MsgBoxStyle.Information, "File Protection")
            Exit Sub
        Else
            Err.Clear()
        End If
        On Error GoTo ErrorHandling

        FileInStream = New IO.FileStream(Path, IO.FileMode.Open, IO.FileAccess.Read, IO.FileShare.Read)
        FileInStream.Read(FileBuffer, 0, InitialBufferSize)

        TempPos = SearchByteArray(FileBuffer, MP3FrameBeginCRC)
        CurrPos = SearchByteArray(FileBuffer, MP3FrameBeginNoCRC)

        If TempPos = 0 Then
            If CurrPos = 0 Then
                Err.Raise(514, "Extractor", "Could not find any MP3 frames in source file.")
            Else
                TempPos = CurrPos
                MP3FrameBegin = MP3FrameBeginNoCRC
            End If
        Else
            MP3FrameBegin = MP3FrameBeginCRC
        End If

        MP3FrameBegin2 = FileBuffer(TempPos + 3)

        CurrPos = SearchByteArray(FileBuffer, ID3v2Begin)

        If CurrPos <> 0 And CurrPos < TempPos Then
            BeginPos = CurrPos
        Else
            BeginPos = TempPos
        End If

        'Determine FirstDWord of extra info

        FrameSize = MP3FrameSize(FileBuffer(TempPos + 2), samplerate)

        Do

            CurrPos = SearchByteArray(FileBuffer, MP3FrameBegin, TempPos + 2)
            If CurrPos = 0 Then
                Err.Raise(517, "Extractor", "Could not find any non-MP3 data after first frame in source file.")
            ElseIf FileBuffer(CurrPos + 3) <> MP3FrameBegin2 Then
                TempPos = CurrPos
            ElseIf CurrPos > (TempPos + FrameSize) Then
                ExtraDataFound = True
                ExtraInfoSize = CurrPos - (TempPos + FrameSize)
            Else
                TempPos = CurrPos
                FrameSize = MP3FrameSize(FileBuffer(TempPos + 2), samplerate)
            End If

        Loop Until ExtraDataFound

        TempPos = BeginPos - ExtraInfoSize

        FirstDWord(0) = FileBuffer(TempPos)
        FirstDWord(1) = FileBuffer(TempPos + 1)
        FirstDWord(2) = FileBuffer(TempPos + 2)
        FirstDWord(3) = FileBuffer(TempPos + 3)

        'Determine ExtraInfoInterval

        CurrPos = TempPos

        Do

            CurrPos = SearchByteArray(FileBuffer, FirstDWord, CurrPos + 4)

            If CurrPos > 0 Then
                ExtraInfoInterval = CurrPos - TempPos
            Else
                Err.Raise(518, "Extractor", "Did not find a pattern in the extra info of the source file.")
            End If

            'Check for 3rd apperance of extra info
        Loop Until CompareByteArrays(FileBuffer, CurrPos + ExtraInfoInterval, FirstDWord)

        FrameSize = ExtraInfoInterval - ExtraInfoSize
        EndOfFile1 = FileSize - FrameSize
        EndOfFile2 = FileSize - ExtraInfoSize

        ReDim FileBuffer(FrameSize - 1)

        FileInStream.Seek(BeginPos, IO.SeekOrigin.Begin)
        FileOutStream = New IO.FileStream(NewPath, IO.FileMode.CreateNew, IO.FileAccess.Write, IO.FileShare.Read)

        Do

            FileInStream.Read(FileBuffer, 0, FrameSize)
            FileOutStream.Write(FileBuffer, 0, FrameSize)
            If FileInStream.Position <= EndOfFile2 Then
                FileInStream.Read(FileBuffer, 0, ExtraInfoSize)
            End If

        Loop While CompareByteArrays(FileBuffer, 0, FirstDWord) And FileInStream.Position <= EndOfFile1

        FileOutStream.Close()
        FileInStream.Close()

        Beep()

        Exit Sub

ErrorHandling:

        MsgBox("Error: #" & Err.Number & vbCrLf & Err.Description, , "Sorry...")
        If Not IsNothing(FileInStream) Then
            FileInStream.Close()
        End If
        If Not IsNothing(FileOutStream) Then
            FileOutStream.Close()
        End If

    End Sub

    Private Function MP3FrameSize(ByVal FrameSizeInfo As Byte, ByRef samplerate As Long) As Long

        Dim bitrate As Long = 0

        Select Case (FrameSizeInfo And &HF0)

            Case &H10
                bitrate = 32000
            Case &H20
                bitrate = 40000
            Case &H30
                bitrate = 48000
            Case &H40
                bitrate = 56000
            Case &H50
                bitrate = 64000
            Case &H60
                bitrate = 80000
            Case &H70
                bitrate = 96000
            Case &H80
                bitrate = 112000
            Case &H90
                bitrate = 128000
            Case &HA0
                bitrate = 160000
            Case &HB0
                bitrate = 192000
            Case &HC0
                bitrate = 224000
            Case &HD0
                bitrate = 256000
            Case &HE0
                bitrate = 320000
            Case Else
                If samplerate = 0 Then
                    Err.Raise(515, "Extractor", "Invalid bitrate in first MP3 frame.")
                Else
                    Err.Raise(519, "Extractor", "Invalid bitrate after first MP3 frame.")
                End If

        End Select

        If samplerate = 0 Then

            Select Case (FrameSizeInfo And &HC)

                Case &H0
                    samplerate = 44100
                Case &H4
                    samplerate = 48000
                Case &H8
                    samplerate = 32000
                Case Else
                    Err.Raise(516, "Extractor", "Invalid sample rate in first MP3 frame.")

            End Select

        End If

        MP3FrameSize = Fix(144 * bitrate / samplerate) + ((FrameSizeInfo And &H2) >> 1)

    End Function

    Private Function SearchByteArray(ByRef LookIn() As Byte, ByRef LookFor() As Byte, Optional ByVal Start As Long = 0) As Long

        Dim LookInLength As Long = UBound(LookIn)
        Dim LookForLength As Long = UBound(LookFor)
        Dim Counter As Long
        Dim SubCounter As Long

        If LookForLength <= LookInLength And Start >= 0 And Start <= LookInLength - LookForLength Then

            For Counter = Start To LookInLength - LookForLength

                If LookIn(Counter) = LookFor(0) Then

                    SearchByteArray = Counter

                    For SubCounter = 1 To LookForLength

                        If LookIn(Counter + SubCounter) <> LookFor(SubCounter) Then
                            SearchByteArray = 0
                            Exit For
                        End If

                    Next

                    If SearchByteArray <> 0 Then Exit Function

                End If

            Next

        End If

    End Function

    Private Function CompareByteArrays(ByRef Array1() As Byte, ByVal Array1Start As Long, ByRef Array2() As Byte) As Boolean

        Dim Array1Length As Long = UBound(Array1)
        Dim Array2Length As Long = UBound(Array2)
        Dim Counter As Long

        If Array2Length <= Array1Length And Array1Start >= 0 And Array1Start <= Array1Length - Array2Length Then

            CompareByteArrays = True

            For Counter = 0 To Array2Length

                If Array1(Counter + Array1Start) <> Array2(Counter) Then
                    CompareByteArrays = False
                    Exit Function
                End If

            Next

        Else

            CompareByteArrays = False

        End If

    End Function

End Module